/*
    FEARLUS model0-7: ParcelCorrectedYieldWeightedCopyingStrategy.m
    Copyright (C) 1999-2003  Macaulay Institute

    This file is part of FEARLUS model0-7, an agent-based model of land use
    change.

    FEARLUS model0-7 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the ParcelCorrectedYieldWeightedCopyingStrategy object. 

*/

#import "ParcelCorrectedYieldWeightedCopyingStrategy.h"
#import "SelectUseBucket.h"
#import "LandUse.h"
#import "LandParcel.h"
#import "LandManager.h"
#import "Environment.h"
#import "Parameter.h"
#import "FearlusArguments.h"
#import "BitString.h"
#import "Interval.h"

@implementation ParcelCorrectedYieldWeightedCopyingStrategy

/*

create:withManager:andParameters:

Create the strategy, setting the manager and parameters.

*/

+(id <ImitativeStrategy, HistoricalStrategy>)create: aZone
	   withManager: (LandManager *)lmgr
	 andParameters: (Parameter *)p {
  ParcelCorrectedYieldWeightedCopyingStrategy *obj;

  obj = [super create: aZone];
  obj->lm = lmgr;
  obj->parameter = p;
  return obj;
}

/*

decideLandUseForParcel:

Decide the land use for the land parcel, using a parcel corrected yield
copying strategy. The score for each parcel considered is yield * match,
with neighbourhood weighting for neighbouring land parcels.

*/

-(LandUse *)decideLandUseForParcel: (LandParcel *)lp {
  SelectUseBucket *bucket;
  id parcelIndex, tempZone;
  LandParcel *myParcel, *neighbouringParcel;
  LandManager *neighbour;
  LandUse *lu;
  int i;
  Interval *yield = [Interval create];
  Interval *match = [Interval create];
  BitString *bs;

  tempZone = [Zone create: scratchZone];
  bucket = [SelectUseBucket create: tempZone
			    withParameters: parameter
			    andLandUses: [[lp getEnvironment] getLandUses]];
  // Loop through land parcels owned by the land manager
  if([FearlusArguments debugCoreMethods]) {
    printf("S: %s: Using parcel corrected yield weighted copying strategy of "
	   "land manager %u to determine land use for LP = land parcel %u at "
	   "(%d, %d)\n", sel_get_name(_cmd), [lm getPIN], [lp getPIN],
	   [lp getX], [lp getY]);
  }
  for(parcelIndex = [[lm getLandParcels] begin: tempZone], [parcelIndex next];
      [parcelIndex getLoc] == Member;
      [parcelIndex next]) {
    myParcel = [parcelIndex get];
    [yield set: [Interval castInt: [myParcel getLastYield]]];
    [match set: [Interval castInt: [ [lp getAttributes]
				     getMatchWith: [myParcel getAttributes] ]
    ]];
    [bucket addScore: [[yield prod: match] call]
	    toBucketForLandUse: [myParcel getLandUse]];
    [myParcel incNImitations];
    if([FearlusArguments debugCoreMethods]) {
      printf("S: %s: Last yield of land parcel %u at (%d, %d) with match to LP"
	     " of %s and land use %u = %s. Score added = %s\n",
	     sel_get_name(_cmd), [myParcel getPIN], [myParcel getX],
	     [myParcel getY], [match printf: "%g"],
	     [[myParcel getLandUse] getPIN], [yield printf: "%g"],
	     [[match prod: yield] printf: "%g"]);
      [Interval printfclean];
    }
  }
  [parcelIndex drop];
  // Loop through the neighbours, and through their land parcels
  bs = [BitString create: tempZone
		  setBitCount: [parameter parcelBitStringSize]];
  for(i = 0; (neighbour = [lm getNextNeighbour: i]); i |= 1) {
    if([FearlusArguments debugCoreMethods]) {
      printf("S: %s: Adding scores for yield of land parcels belonging"
	     " to land manager %u\n", sel_get_name(_cmd), [neighbour getPIN]);
    }
    for(parcelIndex = [[neighbour getLandParcels] begin: tempZone],
	  [parcelIndex next];
	[parcelIndex getLoc] == Member;
	[parcelIndex next]) {
      neighbouringParcel = [parcelIndex get];
      [bs setToCopyOf: [neighbouringParcel getAttributes]];
      [bs mutateWithProbability: [lm getNeighbourNoise]];
      [match set: [Interval castInt: [ [lp getAttributes] getMatchWith: bs ]]];
      [yield set: [Interval castInt: [neighbouringParcel getLastYield]]];
      [bucket addScore:
		[[[[lm getNeighbourWeight] prod: yield] prod: match] call]
	      toBucketForLandUse: [neighbouringParcel getLandUse]];
      [neighbouringParcel incNImitations];
      if([FearlusArguments debugCoreMethods]) {
	printf("S: %s: Yield of land parcel %u at (%d, %d) with match to LP of"
	       " %s and use %u = %s. Score added = %s\n", sel_get_name(_cmd),
	       [neighbouringParcel getPIN], [neighbouringParcel getX],
	       [neighbouringParcel getY], [match printf: "%g"],
	       [[neighbouringParcel getLandUse] getPIN], [yield printf: "%g"],
	       [[[yield prod: [lm getNeighbourWeight]] prod: match]
		 printf: "%g"]);
	[Interval printfclean];
      }
    }
    [parcelIndex drop];
  }
  lu = [bucket getChoice];
  [yield drop];
  [match drop];
  [bucket drop];
  [bs drop];
  [tempZone drop];
  return lu;
}

@end
